(* cw: a Morse code program
   Copyright (C) 2005  Eric C. Cooper <ecc@cmu.edu>
   Released under the GNU General Public License *)

type parameters =
  { morse : Morse.state;
    practice_mode : bool;
    group_size : int;
    alphabet : string }

let show_char c =
  print_char c;
  flush stdout

let realtime_char c =
  Audio.sync ();
  show_char c

let wait_for_user word =
  ignore (read_line ());
  print_endline word;
  flush stdout

let send_word params word =
  Morse.send params.morse word;
  if params.practice_mode then
    wait_for_user word
  else
    Morse.send params.morse "\n"

let send_random_chars params =
  let str = String.create params.group_size in
  let n = String.length params.alphabet in
  let rec loop () =
    for i = 0 to params.group_size - 1 do
      str.[i] <- params.alphabet.[Random.int n]
    done;
    send_word params str;
    loop ()
  in
  Random.self_init ();
  loop ()

(* An alternative to Array.of_list when the length is already known
   and the list must be reversed *)

let rev_array_of_list (list, n) =
  let a = Array.make n "" in
  let rec loop i  = function
    | [] -> ()
    | hd :: tl -> a.(i) <- hd; loop (i-1) tl
  in
  loop (n-1) list;
  a

let with_open_file file f =
  let chan = open_in file in
  try
    let result = f chan in
    close_in chan;
    result
  with e ->
    close_in chan;
    raise e

(* Return an array of items obtained by reading a channel *)

let read_channel read chan =
  (* We use an option type to encode whether or not EOF was seen
     so that the compiler will optimize the tail-recursion properly.
     If we simply write the tail-recursive call with an exception handler,
     the compiler fails to optimize it into iteration. *)
  let rec get_items acc n =
    let opt = try Some (read chan) with End_of_file -> None in
    match opt with
    | None -> (acc, n)
    | Some str -> get_items (str :: acc) (n+1)
  in
  rev_array_of_list (get_items [] 0)

(* Return an array of items obtained by reading a file *)

let read_file ?(read=input_line) file =
  with_open_file file (read_channel read)

let string_fold_right f init str =
  let rec loop i result =
    if i = 0 then result
    else
      let i = i - 1 in
      loop i (f str.[i] result)
  in
  loop (String.length str) init

(* List equivalent of Array.iteri *)

let iteri proc =
  let rec loop i = function
    | x :: rest -> proc i x; loop (i+1) rest
    | [] -> ()
  in
  loop 0

(* Convert between strings and lists of chars *)

let explode string =
  string_fold_right (fun c list -> c :: list) [] string

let implode chars =
  let len = List.length chars in
  let string = String.create len in
  iteri (fun i c -> string.[i] <- c) chars;
  string

let map_string f str = implode (f (explode str))

let morsify = map_string (List.filter Morse.valid_char)

(* (Reversed) list equivalent of Array.init *)

let listi n f =
  let rec loop i result =
    if i < n then loop (i+1) (f i :: result)
    else result
  in
  loop 0 []

let send_random_words params =
  let lines = read_file "/usr/dict/words" in
  let n = Array.length lines in
  let random_word () = morsify lines.(Random.int n) in
  let rec loop () =
    let rec make_group i s =
      if i = params.group_size then s
      else if i = 0 then
	make_group (i+1) (random_word ())
      else
	make_group (i+1) (s ^ " " ^ (random_word ()))
    in
    send_word params (make_group 0 "");
    loop ()
  in
  Random.self_init ();
  loop ()

(* Send random quotes from the Zippy database *)

let send_zippy_lines params =
  let lines = read_file ~read: Zippy.read (Zippy.file ()) in
  (* first item is a comment -- ignore it *)
  let n = Array.length lines - 1 in
  let rec loop () =
    send_word params lines.(1 + Random.int n);
    loop ()
  in
  Random.self_init ();
  loop ()

(* Mutable program parameters with default values.
   These may be changed during command-line argument processing below. *)

let fast_speed = ref 15 (* words per minute *)
let slow_speed = ref 5 (* words per minute *)
let frequency = ref 800 (* Hz *)
let group_size = ref 5 (* characters or words *)
let practice_mode = ref false
let alphabet = ref ""
let random_words = ref false
let yow = ref false
let message = ref ""
let quiet = ref false

let set_fast n =
  fast_speed := n;
  if !slow_speed > n then
    slow_speed := n

let set_slow n =
  slow_speed := n;
  if !fast_speed < n then
    fast_speed := n

let set_tone n =
  frequency := n

let set_group n =
  group_size := n

let random_chars () =
  alphabet := implode (List.filter Morse.valid_char (listi 256 char_of_int))

let set_chars chars =
  alphabet := chars

let do_letters () =
  alphabet := !alphabet ^ "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

let do_numbers () =
  alphabet := !alphabet ^ "0123456789"

let do_punctuation () =
  alphabet := !alphabet ^ ".,?-=:;()/\"$'_+@"

let keywords = [
  ("-f", Arg.Int set_fast, "fast/Farnsworth speed in wpm [15]");
  ("-s", Arg.Int set_slow, "slow speed in wpm [5]");
  ("-t", Arg.Int set_tone, "tone frequency in Hz [800]");
  ("-n", Arg.Int set_group, "size of character or word groups");
  ("-p", Arg.Set practice_mode, "practice mode");
  ("-c", Arg.Unit random_chars, "random characters");
  ("-C", Arg.String set_chars, "specify characters");
  ("-L", Arg.Unit do_letters, "include letters");
  ("-N", Arg.Unit do_numbers, "include numbers");
  ("-P", Arg.Unit do_punctuation, "include punctuation");
  ("-w", Arg.Set random_words, "random words");
  ("-y", Arg.Set yow, "Yow!");
  ("-q", Arg.Set quiet, "quiet mode");
]

let usage_string = "Usage: cw [options] [message ...]"

let usage () =
  Arg.usage keywords usage_string;
  exit 1

let add_arg s =
  if !alphabet <> "" or !random_words or !yow then usage ();
  match !message with
  | "" -> message := s
  | m -> message := m ^ " " ^ s

let () =
  if Array.length Sys.argv = 1 then usage ();
  Arg.parse keywords add_arg usage_string;
  if not !quiet then Audio.init ();
  let send =
    if !quiet then ignore
    else Audio.play
  in
  let display =
    if !practice_mode then ignore
    else if !quiet then show_char
    else realtime_char
  in
  let params =
    { morse = Morse.init ~frequency: !frequency ~fast_wpm: !fast_speed
	~slow_wpm: !slow_speed ~send ~display;
      practice_mode = !practice_mode;
      group_size = !group_size;
      alphabet = !alphabet }
  in
  if !alphabet <> "" then
    send_random_chars params
  else if !random_words then
    send_random_words params
  else if !yow then
    send_zippy_lines params
  else
    send_word params !message
